# =====================================================
#   test_config.py
# --------------
#   Unit test for creating VOevent 
# =====================================================

import unittest
from hek_voevent_tools import load_config, parse_features_file, create_event

class TestCreateEvent(unittest.TestCase):
  def test_create_event(self):
    config = load_config()
    voevent_spec = config['DEFAULT'].get('specs_location')
    csv_data = parse_features_file(voevent_spec)

    # Check event creation
    voevent = create_event(csv_data, 'FL')
    self.assertEqual(voevent.event_type, 'FL')

    # Check for parameters
    voevent.required['FRM_Name'] = 'Flare Detective - Trigger Module'
    voevent.optional['FL_PeakFlux'] = "674.229"
    self.assertEqual(voevent.required['KB_ArchivURL'], 'Reserved for KB archivist: URL to suppl. info.')
    self.assertEqual(voevent.required['FRM_Name'], 'Flare Detective - Trigger Module')
    self.assertEqual(voevent.optional['FL_PeakFlux'], '674.229')


import os
import configparser

def load_config(config_file: str = None):
  config = configparser.ConfigParser()

  # Default is config.ini in same folder
  if config_file is None:  
      config_file = os.path.join(os.path.dirname(__file__), "config.ini")
  if not os.path.exists(config_file):
      raise FileNotFoundError(f"Config file not found: {config_file}")

  config.read(config_file)
  return config


from .voevent import VOevent
from .ingest_csv import parse_features_file
from .create_event import create_event 
from .export_event import Error, export_event, validate_type
from .config import load_config

__all__ = [
  'VOevent',
  'parse_features_file',
  'load_config',
  'create_event',
  'Error',
  'export_event',
  'validate_type',
]




# =====================================================
#   test_config.py
# --------------
#   Unit test for configuration file
# =====================================================

import unittest
from hek_voevent_tools import load_config

class TestConfig(unittest.TestCase):
  def test_load_config(self):
    config = load_config()
    voevent_spec = config['DEFAULT'].get('specs_location')
    output_path = config['DEFAULT'].get('output_path')

    # Check that configurations exists as strings
    self.assertIsInstance(voevent_spec, str)
    self.assertIsInstance(output_path, str)


# =====================================================
#   ingest_specs.py
# --------------
#   Parse VOEvent Spec page
# =====================================================

import io
import requests
import pandas as pd

class Error(Exception):
  pass

# Import CSV and return data object
def parse_features_file(url):
  try:
    response = requests.get(url)
    if response.status_code == 200:
      txt_io = io.StringIO(response.text)
      features = pd.read_csv(txt_io, skiprows=2, header=[0])    
      return features
    else: 
      raise Error('Failed to fetch file. [Status: {response.status_code}]')
  except Exception as e:
    print(f'Failed to load CSV: {e}')
    exit()




# =====================================================
#   test_config.py
# --------------
#   Unit test for ingesting the CSV spec file
# =====================================================

import unittest
import pandas as pd
from hek_voevent_tools import load_config, parse_features_file

class TestIngest(unittest.TestCase):
  def test_ingest_csv(self):
    config = load_config()
    voevent_spec = config['DEFAULT'].get('specs_location')
    csv_data = parse_features_file(voevent_spec)

    # Check that the csv is read
    self.assertIsInstance(csv_data, pd.DataFrame)



# =====================================================
#   Export_event.py
# --------------
#   Generate an XML based on provided VOevent obj
# =====================================================

import xml.etree.ElementTree as etree
from datetime import datetime

class Error(Exception):
  pass

def export_event(df, voevent, output_file): 
  NSMAP = {
    'voe': "http://www.ivoa.net/xml/VOEvent/v1.1",
    'stc': "http://www.ivoa.net/xml/STC/stc-v1.30.xsd",
    'lmsal': "http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd",
    'xsi': "http://www.w3.org/2001/XMLSchema-instance"
  }

  root = etree.Element(
    '{http://www.ivoa.net/xml/VOEvent/v1.1}VOEvent',
    nsmap=NSMAP,
    ivorn=voevent.required['KB_ArchivID'],
    role="observation",
    version="1.1",
    attrib={"{http://www.w3.org/2001/XMLSchema-instance}schemaLocation":
            "http://www.ivoa.net/xml/VOEvent/v1.1 http://www.lmsal.com/helio-informatics/VOEvent-v1.1.xsd"}
  )

  # Create XML structure
  # Get event list dynamically from dataframe
  # - Obtain all abbrevations between Dimensions and Source column
  header = df.columns
  header_top = header.get_level_values(0)
  dimensions_index = list(header_top).index('Dimensions')
  source_index = list(header_top).index('Source')
  source_index = list(header_top).index('Type')
  event_types_abrv = header_top[dimensions_index+1:source_index]
  event_type_names = df.iloc[0, dimensions_index+1:source_index].values
  event_types= dict(zip(event_types_abrv, event_type_names))
  df = df.iloc[1:]  # Remove the top line (Full_Name line)
  print(f'Event type List: {event_types}')

  # Get all VOParamType groups list dynamically from dataframe
  # - What do we do with EventMap_URL with no group?
  # - Currently the '-' group is removed
  vo_type_list = df[('VOParamType')].unique()
  vo_type_list = list(filter(lambda x: x != '-', vo_type_list)) 
  print(f'VOParamType list: {vo_type_list}')

  
  # Object Validation
  try:
    # Validate that all parameters in required / optional exist according to .csv
    # ---------------------
    combined_dict = {**voevent.required, **voevent.optional}
    combined_dict_keys = list(combined_dict.keys())
    not_matched = [item for item in combined_dict_keys if item not in df['Parameter'].values]
    if len(not_matched)>0: raise Error(f'Unknown attributes found - {not_matched}')

    # Requirements validation (as export_event.pro has it)
    # ---------------------
    validate, err_list = True, [] 
    for key, value in voevent.required.items():
      if value == "" or None:
        print(f"Required parameter '[{key}]' value is missing!")
        err_list.append(key)
    if validate == False: raise Error(f'Required keys {err_list} have no values!')
    print('All voevent.required values filled')

    # Type check validation (as export_event.pro has it)
    # ---------------------
    validate, err_list = True, []
    for key, value in combined_dict.items():
      value_type = df[df['Parameter'] == key].iloc[0]['Type']
      if key == 'Event_CoordSys':
        if str(value).lower() not in ['utc-hpc-topo', 'utc-hpr-topo', 'utc-hgs-topo', 'utc-hgc-topo', 'utc-hcr-topo']:
          validate = False
          err_list.append(key)
          print(f"Invalid value for '{key}': {value}!")
          print('''---------------------------Error--------------------------
          Please use one of the following coordinate systems
          for the required.coord_sys attribute.
          Cartesian helio-projective:      "utc-hpc-topo"
          Polar helio-projective:          "utc-hpr-topo"
          Stonyhurst heliographic:         "utc-hgs-topo"
          Carrington heliographic:         "utc-hgc-topo"
          Pol. angle and radius (for CMEs):"utc-hcr-topo"
          ----------------------------------------------------------
          For details, go to:
          http://www.ivoa.net/Documents/PR/VOE/VOEvent-20060629.html
          ----------------------------------------------------------''')
        else:
          print(f"Valid value for '{key}': {value} (string / Event_CoordSys)")
      elif value is not "" or None:
        if validate_type(value, value_type):
          print(f"Valid value for '{key}': {value} ({value_type})")
        else:
          validate = False
          err_list.append(key)
          print(f"Invalid value for '{key}': {value} (Expected: {value_type})!")
    if validate == False: raise Error(f'Keys {err_list} have invalid value types!')
  except Error as e:
    print(f'ERROR: {e}')
    exit()
  print('All voevent.required and voevent.optional values types are checked')

  # Required for ALL types... Manually constructed tree (r in R/O column)
  # ---------------------
  # Who
  Who = etree.SubElement(root, "Who")
  Who.append(etree.Comment(" Data pertaining to curation "))
  etree.SubElement(Who, "AuthorIVORN").text = voevent.required['KB_ArchivURL']
  Author = etree.SubElement(Who, "Author")
  etree.SubElement(Author, "contactName").text = voevent.required['KB_Archivist']
  etree.SubElement(Who, "Date").text = voevent.required['KB_ArchivDate']

  # What
  What = etree.SubElement(root, "What")
  What.append(etree.Comment(" Data about what was measured/observed. "))

  # WhereWhen
  WhereWhen = etree.SubElement(root, "WhereWhen")
  WhereWhen.append(etree.Comment(" Data pertaining to when and where something occured "))
  ObsDataLocation = etree.SubElement(WhereWhen, "ObsDataLocation", xmlns="http://www.ivoa.net/xml/STC/stc-v1.30.xsd")

  ObservatoryLocation = etree.SubElement(ObsDataLocation, "ObservatoryLocation")
  etree.SubElement(ObservatoryLocation, "AstroCoordSystem")
  etree.SubElement(ObservatoryLocation, "AstroCoords", id=voevent.required['Event_CoordSys'], coord_system_id=voevent.required['Event_CoordSys'])

  ObservationLocation = etree.SubElement(ObsDataLocation, "ObservationLocation", id=voevent.required['OBS_Observatory'])
  etree.SubElement(ObservationLocation, "AstroCoordSystem")

  AstroCoords_Obs = etree.SubElement(ObservationLocation, "AstroCoords", coord_system_id=voevent.required['Event_CoordSys'])

  Time = etree.SubElement(AstroCoords_Obs, "Time")
  TimeInstant = etree.SubElement(Time, "TimeInstant")
  etree.SubElement(TimeInstant, "ISOTime").text = voevent.required['Event_StartTime']

  Position2D = etree.SubElement(AstroCoords_Obs, "Position2D", unit=voevent.required['Event_CoordUnit'])
  Value2 = etree.SubElement(Position2D, "Value2")
  etree.SubElement(Value2, "C1").text = voevent.required['Event_Coord1']
  etree.SubElement(Value2, "C2").text = voevent.required['Event_Coord2']
  Error2 = etree.SubElement(Position2D, "Error2")
  etree.SubElement(Error2, "C1").text = voevent.required['Event_C1Error']
  etree.SubElement(Error2, "C2").text = voevent.required['Event_C2Error']

  AstroCoordArea = etree.SubElement(ObservationLocation, "AstroCoordArea", coord_system_id=voevent.required['Event_CoordSys'])
  TimeInterval = etree.SubElement(AstroCoordArea, "TimeInterval")
  StartTime = etree.SubElement(TimeInterval, "StartTime")
  etree.SubElement(StartTime, "ISOTime").text = voevent.required['Event_StartTime']
  StopTime = etree.SubElement(TimeInterval, "StopTime")
  etree.SubElement(StopTime, "ISOTime").text = voevent.required['Event_EndTime']
  
  # Need to manually calculate size of box from given parameters
  Box2 = etree.SubElement(AstroCoordArea, "Box2")
  Center = etree.SubElement(Box2, "Center")
  etree.SubElement(Center, "C1").text = voevent.required['Event_Coord1']
  etree.SubElement(Center, "C2").text = voevent.required['Event_Coord2']
  Size = etree.SubElement(Box2, "Size")
  etree.SubElement(Size, "C1").text = str("{:.5f}".format(float(voevent.required['BoundBox_C1UR']) - float(voevent.required['BoundBox_C1LL'])))
  etree.SubElement(Size, "C2").text = str("{:.5f}".format(float(voevent.required['BoundBox_C2UR']) - float(voevent.required['BoundBox_C2LL'])))

  # How
  How = etree.SubElement(root, "How")
  How.append(etree.Comment("  Data pertaining to how the feature/event detection was performed  "))
  lmsal_data = etree.SubElement(How, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}data")
  etree.SubElement(lmsal_data, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}OBS_ChannelID").text = voevent.required['OBS_ChannelID']
  etree.SubElement(lmsal_data, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}OBS_Instrument").text = voevent.required['OBS_Instrument'] 
  etree.SubElement(lmsal_data, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}OBS_MeanWavel").text = voevent.required['OBS_MeanWavel']
  etree.SubElement(lmsal_data, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}OBS_WavelUnit").text = voevent.required['OBS_WavelUnit']
  lmsal_method = etree.SubElement(How, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}method")
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_Contact").text = voevent.required['FRM_Contact']
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_DateRun").text = voevent.required['FRM_DateRun']
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_HumanFlag").text = voevent.required['FRM_HumanFlag']
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_Identifier").text = voevent.required['FRM_Identifier']
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_Institute").text = voevent.required['FRM_Institute']
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_Name").text = voevent.required['FRM_Name']
  etree.SubElement(lmsal_method, "{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_ParamSet").text = voevent.required['FRM_ParamSet']

  # Why
  Why = etree.SubElement(root, "Why")
  # Extra case for event probability
  if "Event_Probability" in voevent.optional and voevent.optional["Event_Probability"] is not "":
    etree.SubElement(Why, "Param", name="Inference", probability=voevent.optional["Event_Probability"])
  # etree.SubElement(Why, "Inference", probability="Inf")
  etree.SubElement(Why, "Concept").text = voevent.event_fullname
  etree.SubElement(Why, "EVENT_TYPE").text = voevent.required['Event_Type']
  if voevent.description: etree.SubElement(Why, "Description").text = voevent.description 

  # Map string to branches
  BRANCHES = {
    'who': Who,
    'what': What,
    'wherewhen': WhereWhen,
    'why': Why,
    'how': How
  }

  # Required for specified event type (o in R/O column)
  # ---------------------
  # Group = etree.SubElement(WhereWhen, "Group", name="Flare_required")
  # etree.SubElement(Group, "Param", name="EVENT_PEAKTIME", value=f"{voevent.required['Event_PeakTime']}")
  for key, value in voevent.required.items():
    row = df[df['Parameter'] == key].iloc[0]
    if row['R/O'] == 'r': continue
    branch, group_name  = BRANCHES[row['VOParamType']], f"{voevent.event_fullname}_required"
    group = branch.find(f"./Group[@name='{group_name}']")
    if group is None: group = etree.SubElement(branch, "Group", name=group_name)
    etree.SubElement(group, "Param", name=row['Parameter'].upper(), value=value)

  # Fully Optional (o in R/O column)
  # ---------------------
  for key, value in voevent.optional.items():
    if value == "": continue
    row = df[df['Parameter'] == key].iloc[0]
    branch, group_name  = BRANCHES[row['VOParamType']], f"{voevent.event_fullname}_optional"
    group = branch.find(f"./Group[@name='{group_name}']")
    if group is None: group = etree.SubElement(branch, "Group", name=group_name)
    etree.SubElement(group, "Param", name=row['Parameter'].upper(), value=value)

  # Reference
  # ---------------------
  etree.SubElement(root, "Reference", name='FRM_URL', uri=voevent.required['FRM_URL'])
  for ref in voevent.reference:
    etree.SubElement(root, "Reference", name=ref['name'], type=ref['type'], link=ref['link'])

  # Citation
  if voevent.citation:
    if voevent.citation['action'] not in ['action', 'supersedes', 'retraction']:
      print('ERROR: Citation action must be: followup / supersedes / retraction')
      exit()
    Citations = etree.SubElement(root, "Citations")
    etree.SubElement(Citations, "EventIVORN", action=voevent.citation['action']).text = voevent.required['KB_ArchivID']
    etree.SubElement(Citations, "Description").text = voevent.citation['description']

  # Export
  # ---------------------
  # Print as string
  # xml_bytes = etree.tostring(root, pretty_print=True, xml_declaration=True, encoding='UTF-8')
  # print(xml_bytes.decode('utf-8'))

  # Write out to file
  etree.indent(root, space="  ", level=0)
  tree = etree.ElementTree(root)
  tree.write(output_file, encoding="utf-8", xml_declaration=True)
  print(f'Finished! Exported XML to [{output_file}]')
  print()

# Define the function to validate the value type
def validate_type(value, type):
  date_formats = [
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%dT%H:%M:%S.%f" # ISO but with milliseconds
  ]
  match type:
    case 'float':  
      try: 
        float(value) 
        return True
      except ValueError: return False
    case 'string': 
      try: 
        str(value) 
        return True
      except ValueError: return False 
    case 'integer' | 'long': # In Python 3, long is equivalent to int
      try: 
        int(value) 
        return True
      except ValueError: return False 
    case 'boolean': return str(value).lower() in ['true', 'false', 't', 'f']
    case 'polygon': return True  
    case 'line': return True
    case 'datetime':
      for fmt in date_formats:
        try:
          datetime.strptime(value, fmt)
          return True
        except ValueError: continue
      return False
    case _: return False


# =====================================================
#   voevent.py
# --------------
#   Class definition for VOEvent
# =====================================================

class VOevent:
  event_type: str
  event_fullname: str
  description: str

  # First time initialization - generate attributes from given lists
  def __init__(self, event_type, required_list, optional_list):
    self.event_type = event_type['abrv']
    self.event_fullname = event_type['full']
    self.required = {key: "" for key in required_list}
    self.optional = {key: "" for key in optional_list}
    self.description = None
    self.citation = []
    self.reference = []

    self.required['Event_Type'] = f"{event_type['abrv']}: {event_type['full']}"
    self.required['KB_ArchivID'] = 'Reserved for KB archivist: KB entry identifier'
    self.required['KB_Archivist'] = 'Reserved for KB archivist: KB entry made by'
    self.required['KB_ArchivDate'] = 'Reserved for KB archivist: KB entry date'
    self.required['KB_ArchivURL'] = 'Reserved for KB archivist: URL to suppl. info.'
    self.required['Event_StartTime'] = '1492-10-12 00:00:00'
    self.required['Event_EndTime'] = '1492-10-12 00:00:00'
    # self.required['Event_Expires'] = '1492-10-12 00:00:00'

  # Display the supported feature list

  # Display details of attributes
  def help(self):
    print('==============================')
    print('Required Attributes: ')
    print('-------------------- ')
    for key, value in self.required.items(): print(f'{key}: {value}')
    print()

    print('Optional Attributes: ')
    print('-------------------- ')
    for key, value in self.optional.items(): print(f'{key}: {value}')
    print()
    print("For further info, go to http://www.lmsal.com/helio-informatics/hpkb/VOEvent_Spec.html")
    print('==============================')

  # String representation of class 
  def __repr__(self):
    return f'''VOevent [{self.event_type}]
    
---------------
required={self.required}

optional={self.optional}
'''
    # return f"VOevent(type={self.event_type}\nrequired={self.required}\noptional={self.optional})"


# =====================================================
#   test_config.py
# --------------
#   Unit test for checking XML output
# =====================================================

import unittest
import tempfile
import os
from datetime import datetime
import xml.etree.ElementTree as etree
from hek_voevent_tools import load_config, parse_features_file, create_event, export_event

class TestExportXML(unittest.TestCase):
  def test_export_xml(self):
    # Make the output .xml a temporary file
    with tempfile.NamedTemporaryFile(delete=False, suffix=".xml") as temp_file:
      temp_path = temp_file.name

    # Make object
    config = load_config()
    csv_data = parse_features_file(config['DEFAULT'].get('specs_location'))
    voevent = create_event(csv_data, 'FL')

    voevent.required['KB_ArchivURL'] = 'ivo://helio-informatics.org/FL_FlareDetective-TriggerModule_20250703_143416_2025-07-03T08:50:05.835_1' + "_test1"
    voevent.required['KB_ArchivID'] = 'ivo://helio-informatics.org/FL_FlareDetective-TriggerModule_20250703_143416_2025-07-03T08:50:05.835_1' + "_test1"
    voevent.required['KB_Archivist'] = 'Paolo C. Grigis - pgrigis@cfa.harvard.edu'
    voevent.required['KB_ArchivDate'] = datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S")
    voevent.required['OBS_Observatory'] = 'SDO'
    voevent.required['OBS_Instrument'] = 'AIA'
    voevent.required['OBS_ChannelID'] = '193'
    voevent.required['OBS_MeanWavel'] = '193.000'
    voevent.required['OBS_WavelUnit'] = 'angstrom'
    voevent.required['FRM_Name'] = 'Flare Detective - Trigger Module'
    voevent.required['FRM_Identifier'] = 'Feature Finding Team'
    voevent.required['FRM_Institute'] = 'SAO'
    voevent.required['FRM_HumanFlag'] = 'F'
    voevent.required['FRM_ParamSet'] = 'DerivativeThreshold= 3.00000e+00 EndFraction= 2.50000e-01'
    voevent.required['FRM_DateRun'] = '2025-07-03T07:34:13.000'
    voevent.required['FRM_Contact'] = 'Paolo C. Grigis - pgrigis@cfa.harvard.edu'
    voevent.required['FRM_URL'] = 'http://www.cfa.harvard.edu'
    voevent.optional['FRM_VersionNumber'] = "0.510000"
    voevent.required['Event_StartTime'] = '2025-07-03T08:49:05.835'
    voevent.required['Event_PeakTime'] = '2025-07-03T08:50:17.846'
    voevent.required['Event_EndTime'] = '2025-07-03T08:51:53.835'
    voevent.required['Event_CoordSys'] = 'UTC-HPC-TOPO'
    voevent.required['Event_CoordUnit'] = 'arcseconds'
    voevent.required['Event_Coord1'] = '-576.000'
    voevent.required['Event_Coord2'] = '-345.600'
    voevent.required['Event_C1Error'] = '2.00000'
    voevent.required['Event_C2Error'] = '2.00000'
    voevent.required['BoundBox_C1LL'] = '-440' #Coordinates of lower-left
    voevent.required['BoundBox_C2LL'] = '263.2'  #Corner of bounding box
    voevent.required['BoundBox_C1UR'] = '-363.2' #Coordinates of upper-right    
    voevent.required['BoundBox_C2UR'] = '340'  #Corner of bounding box 
    voevent.optional['FL_PeakFlux'] = "674.229"
    voevent.optional['FL_PeakFluxUnit'] = "DN/sec/pixel"
    voevent.optional['Event_TestFlag'] = 'True'
    voevent.citation = {
      'action': 'supersedes',
      'description': 'Closing of open event'
    }
    voevent.reference.append({
      'name': 'Publication',
      'link': 'http://www.harvard.edu/',
      'type': 'html'
    })
    export_event(csv_data, voevent, temp_path)

    # Check that file exists
    self.assertTrue(os.path.exists(temp_path))

    # Check xml contents
    tree = etree.parse(temp_path)
    root = tree.getroot()
    what = root.find('.//What') 
    fl_peakflux = what.find('.//Param[@name="FL_PEAKFLUX"]').attrib['value'] 
    how = root.find('.//How') 
    frm_name = how.find('.//{http://www.lmsal.com/helio-informatics/lmsal-v1.0.xsd}FRM_Name').text 
    self.assertEqual(fl_peakflux, "674.229")
    self.assertEqual(frm_name, "Flare Detective - Trigger Module")

    # Remove the temporary xml
    os.remove(temp_path)


# =====================================================
#   create_event.py
# --------------
#   Create the Element Tree XML object
#   - Based on struct4event.pro
# =====================================================

from .voevent import VOevent

# Generate VOevent obj
def create_event(df, event_type_abrv):
  # Get event list dynamically from dataframe
  # - Obtain all abbrevations between Dimensions and Source column
  header = df.columns
  header_top = header.get_level_values(0)
  dimensions_index = list(header_top).index('Dimensions')
  source_index = list(header_top).index('Source')
  event_types_abrv = header_top[dimensions_index+1:source_index]
  event_type_names = df.iloc[0, dimensions_index+1:source_index].values
  event_types= dict(zip(event_types_abrv, event_type_names))
  df = df.iloc[1:]  # Remove the top line (Full_Name line)
  # print(f'Event type List: {event_types}')

  # Make full event_type string
  event_type={}
  event_type['abrv'], event_type['full'] = event_type_abrv, event_types[event_type_abrv]
  print(event_type['full'])

  # Given the event type, get the list of required and optional parameters'
  required_attr, optional_attr = [], []
  for index, row in df.iterrows():
    parameter, dimensions, attribute_type, source, vo_param_type, r_o, vo_translation, description =  \
      row['Parameter'], row['Dimensions'], row['Type'], row['Source'], row['VOParamType'], row['R/O'], row['VOTranslation'], row['Description']
    required = row.iloc[3+list(event_types.keys()).index(event_type_abrv)]
    # print(f'{input_event_type}, {parameter}, {attribute_type}, {required}, {source}, {vo_param_type}, {r_o}')
    
    # Check if the attribute is required or optional for the specified event type
    match int(required):
      case 9: 
        required_attr.append(parameter)
      case 5: 
        optional_attr.append(parameter)

  # Generate the structure with the required / optional parameters
  voevent_obj = VOevent(event_type, required_attr, optional_attr)

  # Finished - Return the structure
  return voevent_obj





  # # Check if the attribute is required or optional for the specified event type
  # if input_event_type in event_types and required != 0: 
  #   for sub_elem in root:
  #     if sub_elem.tag == vo_param_type.capitalize() \
  #     or (vo_param_type == 'wherewhen' and sub_elem.tag == 'WhereWhen'):
  #       if r_o == 'r':        
  #         match source:
  #           case 'data': ET.SubElement(sub_elem.find('lmsal:data'), parameter)
  #           case 'method': ET.SubElement(sub_elem.find('lmsal:method'), parameter)
  #           case _: ET.SubElement(sub_elem, parameter)
  #       else: 
  #         group_elem = sub_elem.find('Group')
  #         set_optional_param(group_elem, parameter, '')
  #       # elem.text = ''  # Placeholder, replace with actual attribute value
 
  
  # # Dynamically generate different VOParamType groups and associated branches
  # root = ET.Element("voe:VOEvent")
  # for vo_type in vo_type_list:
  #   if vo_type == 'wherewhen': 
  #     sub_elem = ET.SubElement(root, 'WhereWhen')
  #   else: 
  #     sub_elem = ET.SubElement(root, vo_type.capitalize())
  #   ET.SubElement(sub_elem, "lmsal:data")
  #   ET.SubElement(sub_elem, "lmsal:method")
  #   group_elem = ET.SubElement(sub_elem, "Group")
  #   group_elem.set('Name', 'Other_optional')

  # # Iterate through entire dataframe
  # # - Find attributes for the specified event type
  # for index, row in df.iterrows():
  #   parameter, attribute_type, source, vo_param_type, r_o =  \
  #     row['Parameter'], row['Type'], row['Source'], row['VOParamType'], row['R/O']
  #   required = row.iloc[3+list(event_types.keys()).index(input_event_type)]
  #   print(f'{input_event_type}, {parameter}, {attribute_type}, {required}, {source}, {vo_param_type}, {r_o}')
    
  #   # Check if the attribute is required or optional for the specified event type
  #   if input_event_type in event_types and required != 0: 
  #     for sub_elem in root:
  #       if sub_elem.tag == vo_param_type.capitalize() \
  #       or (vo_param_type == 'wherewhen' and sub_elem.tag == 'WhereWhen'):
  #         if r_o == 'r':        
  #           match source:
  #             case 'data': ET.SubElement(sub_elem.find('lmsal:data'), parameter)
  #             case 'method': ET.SubElement(sub_elem.find('lmsal:method'), parameter)
  #             case _: ET.SubElement(sub_elem, parameter)
  #         else: 
  #           group_elem = sub_elem.find('Group')
  #           set_optional_param(group_elem, parameter, '')
  #         # elem.text = ''  # Placeholder, replace with actual attribute value
  
  # # Finished - Return tree structure
  # return root